{
Copyright 2005 Darwin O'Connor

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
program captrec;
{$PMType PM}
uses sysutils, VPUtils, os2def, os2base, os2mm, os2pmapi, dart;

type
   EMMIOError = Exception;
   tbitmap = array[0..10000000] of byte;
   pbitmap = ^tbitmap;
   EMCIErr = class(Exception)
   public
      rc : ULONG;
      constructor CreateMCIErr(inrc : ULONG);
   end;

var
   ab : Hab;
   mq : Hmq;
   Frame,
   Client : HWND;
   wavmmio,
   mmio : hmmio;
   log : text;

const
   factor = 1000.0*60.0*60.0*24.0;
   MaxBuffers = 30;

var
   LengthSec : integer;
   basename : string;
   filenum : integer;
   framecount : integer;
   Quality : ULONG;
   framerate : double;
   framelen : double;
   width,
   Height : integer;
   CODECDLL,
   CODECEntry : string;
   DARTDevice : USHORT;
   BufferList : array[0..MaxBuffers-1] of MCI_MIX_BUFFER;
   BufferCount : ULONG;
   DartRead : MixerProc;
   BuffersReady,
   CurrBuffer : integer;
   MixHandle : ULONG;
   Sem : HEV;
   blocking : boolean;
   AudioBufferSize : integer;
   AudioBytes : integer;
   AudioRate : integer;
   Channels : integer;
   HasAudio : boolean;
   WindowTitle : string;
   bitcount : integer;
   BuffReadyMutex : HMTX;
   BufferFraction : double;
   SampleSize : integer;
   ThreadId : longInt;
   AbortRec : boolean;
   BttvFrame,
   BttvClient : HWND;
   UnCompress : boolean;
   NoFlip : boolean;
   FrameWindow : boolean;

constructor EMCIErr.CreateMCIErr(inrc : ULONG);
const
   MaxLen = 128;
var
   ErrString : string;
   Loop : integer;
begin
   rc:=inrc;
   SetLength(ErrString,MaxLen);
   mciGetErrorString(rc,pchar(ErrString),MaxLen);
   loop:=1;
   while (loop<=MaxLen) and (ErrString[loop]<>#0) do Inc(loop);
   SetLength(ErrString,loop);
   inherited Create(ErrString);
end;

procedure MMCheck(rc : ULong);
begin
   if (rc and $FFFF)<>MCIERR_SUCCESS then raise EMCIErr.CreateMCIErr(rc);
end;

procedure mmioCheck(rc: ULONG; s : string);
begin
   if rc<>MMIO_SUCCESS then begin
      if rc=MMIO_ERROR then rc:=mmioGetLastError(mmio);
      raise EMMIOError.Create('MMIO Error: '+IntToStr(rc-MMIOERR_BASE)+' ('+IntToHex(rc,8)+'h) in '+s);
   end;
end;

procedure SetActiveTrack(mmio : hmmio; Track : ULONG);
var
   ExtInfo : MMEXTENDINFO;
begin
   fillChar(ExtInfo,SizeOf(ExtInfo),0);
   ExtInfo.ulFlags:=MMIO_TRACK;
   ExtInfo.ulTrackId:=Track;
   mmioCheck(mmioSet(mmio,@ExtInfo,MMIO_SET_EXTENDEDINFO),'mmioSet');
end;

procedure OpenFile;
type
   TVideoHead = record
      h : MMVIDEOHEADER;
      a,
      b : ULONG
   end;
var
   mymmioInfo : MMIOINFO;
   MovieHeader : MMMOVIEHEADER;
   VideoHeader : TVIDEOHEAD;
   SourceCODEC,
   DestCODEC : CODECVIDEOHEADER;
   VideoOpen : MMVIDEOOPEN;
   MyCODECOpen : CodecOpen;
   MyCODECAssoc : CODECAssoc;
   ExtendInfo : MMEXTENDINFO;
   Wrote : ULONG;
   CODECInfo : CODECIniFileInfo;
   CompressInfo : MMCOMPRESS;
   VideoCompressInfo : MMVIDEOCOMPRESS;
   Vid : RECORDTABWRITE;
   TrackInfo : array[0..1] of MMTrackInfo;
   filename : string;
   AudioHeader : MMAUDIOHEADER;
begin
   fillChar(mymmioInfo,sizeof(MyMMIOInfo),0);
   with MyMMIOInfo do begin
      ulFlags:=MMIO_READWRITE+MMIO_CREATE;
      ulTranslate:=MMIO_TRANSLATEHEADER OR MMIO_TRANSLATEDATA;
      aulInfo[3]:=MMIO_MEDIATYPE_MOVIE;
      fccIOProc:=mmioFourCC('A','V','I',' ');
//         pIOProc:=mmioInstallIOProc(mmioFourCC('A','V','I',' '),0,MMIO_FINDPROC);
   end;
   if filenum=1 then
      filename:=basename+'.AVI'
   else if filenum>=10 then
      filename:=basename+inttostr(filenum)+'.AVI'
   else
      filename:=basename+'0'+inttostr(filenum)+'.AVI';
   mmio:=mmioOpen(pchar(filename),@mymmioinfo,MyMMIOInfo.ulFlags);
   fillchar(MovieHeader,sizeof(MovieHeader),0);
   with MovieHeader do begin
      ulStructLen:=sizeof(MovieHeader);
      ulMediaType:=MMIO_MEDIATYPE_MOVIE;
      ulMovieCapsFlags:=MOVIE_HAS_VIDEO or MOVIE_CAN_SEEK or MOVIE_CAN_SCAN;
      ulPaddingGranularity:=0;
      ulSuggestedBufferSize:=bitcount*width*Height div 8*2;
      ulLength:=FrameCount;
      if HasAudio then begin
         ulNumEntries:=2;
         ulMovieCapsFlags:=ulMovieCapsFlags or MOVIE_HAS_AUDIO;
      end else ulNumEntries:=1;
      pmmTrackInfoList:=@TrackInfo;
   end;
   fillchar(TrackInfo,sizeof(TrackInfo),0);
   with TrackInfo[0] do begin
      ulMediaType:=MMIO_MEDIATYPE_DIGITALVIDEO;
   end;
   TrackInfo[1].ulTrackID:=1;
   TrackInfo[1].ulMediaType:=MMIO_MEDIATYPE_AUDIO;
   SetActiveTrack(mmio,MMIO_RESETTRACKS);
   mmioCheck(mmioSetHeader(mmio,@MovieHeader,sizeof(MovieHeader),@Wrote,0,0),'mmioSetHeader main');
   SetActiveTrack(mmio,0);
//      mmioCheck(mmioQueryHeaderLength(mmio,@Wrote,0,0),'mmioQueryHeaderLength');
   fillchar(VideoHeader,sizeof(VideoHeader),0);
   with VideoHeader.h do begin
      ulStructLen:=sizeof(VideoHeader.h);
      ulContentType:=MMIO_VIDEO_DATA;
      ulMediaType:=MMIO_MEDIATYPE_DIGITALVIDEO;
      ulWidth:=width;
      ulHeight:=Height;
      ulScale:=100;
      ulRate:=round(100*framerate);
      mmtimePerFrame:=round(3000/framerate);
   end;
   mmioCheck(mmioSetHeader(mmio,@VideoHeader,sizeof(videoheader),@wrote,0,0),'mmioSetHeader video');
   fillchar(AudioHeader,sizeof(AudioHeader),0);
   with AudioHeader do begin
      ulHeaderLength:=sizeof(AudioHeader);
      ulContentType:=MMIO_AUDIO_MUSIC;
      ulMediaType:=MMIO_MEDIATYPE_AUDIO;
      with mmXWAVHeader.WaveHeader do begin
         usFormatTag:=DATATYPE_WAVEFORM;
         usChannels:=Channels;
         ulSamplesPerSec:=AudioRate;
         ulAvgBytesPerSec:=AudioRate*SampleSize;
         usBlockAlign:=SampleSize;
         usBitsPerSample:=AudioBytes*8;
      end;
   end;
   if HasAudio then begin {AudioHeader saved for wavmmio if not HasAudio}
      SetActiveTrack(mmio,1);
      mmioCheck(mmioSetHeader(mmio,@AudioHeader,sizeof(Audioheader),@wrote,0,0),'mmioSetHeader audio');
   end;
   SetActiveTrack(mmio,MMIO_RESETTRACKS);
   fillchar(SourceCODEC,sizeof(SourceCODEC),0);
   with SourceCODEC do begin
      ulStructLen:=sizeof(SourceCODEC);
      cx:=width;
      cy:=Height;
      cPlanes:=1;
      cBitCount:=bitcount;
      if bitcount=24 then ulColorEncoding:=MMIO_RGB_24
      else if bitcount=16 then ulColorEncoding:=MMIO_RGB_5_6_5;
   end;
   fillchar(DestCODEC,sizeof(DestCODEC),0);
   with DestCODEC do begin
      ulStructLen:=sizeof(DestCODEC);
      cx:=width;
      cy:=Height;
      cPlanes:=1;
      cBitCount:=bitcount;
      ulColorEncoding:=MMIO_COMPRESSED;
{      if bitcount=24 then ulColorEncoding:=MMIO_RGB_24
      else if bitcount=16 then ulColorEncoding:=MMIO_RGB_5_6_5;}
   end;
   with VideoOpen do begin
      ulStructLen:=sizeof(CODECOpen);
      ulQuality:=Quality;
      ulKeyFrameRate:=-1;
      ulScale:=100;
      ulRate:=round(100*Framerate);
      ulDataConstraint:=-1;
      ulConstraintInterval:=-1;
   end;
   fillchar(MyCodecOpen,sizeof(CODECOpen),0);
   with MyCodecOpen do begin
      ulFlags:=CODEC_COMPRESS;
      pControlHdr:=nil;
      pSrcHdr:=@SourceCODEC;
      pDstHdr:=@DestCODEC;
      pOtherInfo:=@VideoOpen;
   end;
   MyCodecAssoc.pCodecOpen:=@MyCodecOpen;
   FillChar(CODECInfo,SizeOf(CODECInfo),0);
   CODECINFO.ulStructLen:=SizeOf(CODECInfo);
   strcopy(@(CODECInfo.szDLLName),pchar(CODECDLL));
   strcopy(@(CODECInfo.szProcName),pchar(CODECEntry));
   mmioCheck(mmioIniFileCODEC(@CODECInfo,MMIO_FINDPROC or MMIO_MATCHDLL or MMIO_MATCHPROCEDURENAME),'mmioIniFileCODEC');
   MyCodecAssoc.pCODECIniInfo:=@CodecInfo;
   fillchar(ExtendInfo,sizeof(ExtendInfo),0);
   with ExtendInfo do begin
      ulStructLen:=sizeof(ExtendInfo);
      ulFlags:=MMIO_CODEC_ASSOC;
      ulNumCODECs:=1;
      pCODECAssoc:=@MyCODECAssoc;
   end;
   mmioCheck(mmioSendMessage(mmio,MMIOM_SET,lONG(@ExtendInfo),MMIO_SET_EXTENDEDINFO),'MMIOM_SET');
   if not HasAudio then begin
      filename:=copy(filename,1,length(filename)-3)+'WAV';
      fillChar(mymmioInfo,sizeof(MyMMIOInfo),0);
      with MyMMIOInfo do begin
         ulFlags:=MMIO_READWRITE+MMIO_CREATE;
         ulTranslate:=MMIO_TRANSLATEHEADER OR MMIO_TRANSLATEDATA;
         fccIOProc:=mmioFourCC('W','A','V','E');
   //         pIOProc:=mmioInstallIOProc(mmioFourCC('A','V','I',' '),0,MMIO_FINDPROC);
      end;
      wavmmio:=mmioOpen(pchar(filename),@mymmioinfo,MyMMIOInfo.ulFlags);
      if wavmmio=0 then mmioCheck(mymmioinfo.ulErrorRet,'mmioOpen audio');
      mmioCheck(mmioSetHeader(wavmmio,@AudioHeader,sizeof(Audioheader),@wrote,0,0),'mmioSetHeader audio');
   end;
end;

function Event(ulStatus : ULONG; var Buffer : MCI_MIX_BUFFER; ulFlags : ULONG) : LONG; cdecl;
begin
   if (ulFlags and MIX_READ_COMPLETE)<>0 then begin
      DosRequestMutexSem(BuffReadyMutex,-1);
      inc(BuffersReady);
      DosReleaseMutexSem(BuffReadyMutex);
   //   writeln('Reading ',BuffersReady);
      if blocking then begin
         DosPostEventSem(sem);
         blocking:=false;
      end;
   end;
end;

procedure OpenDart;
var
   OpenParms : MCI_AMP_OPEN_PARMS;
   SetupParms : MCI_MIXSETUP_PARMS;
   BufferParms : MCI_BUFFER_PARMS;
   ConnectParms : MCI_CONNECTOR_PARMS;
begin
   fillchar(OpenParms,sizeof(OpenParms),0);
   OpenParms.pszDeviceType:=pchar(MCI_DEVTYPE_AUDIO_AMPMIX);
   MMCheck(mciSendCommand(0,MCI_OPEN,MCI_WAIT+MCI_OPEN_TYPE_ID,OpenParms,0));
   DartDevice:=OpenParms.usDeviceID;
   fillChar(ConnectParms,sizeof(ConnectParms),0);
   ConnectParms.ulConnectorType:=MCI_LINE_IN_CONNECTOR;
   MMCheck(mciSendCommand(DartDevice,MCI_CONNECTOR,MCI_WAIT+MCI_ENABLE_CONNECTOR+MCI_CONNECTOR_TYPE,
   ConnectParms,0));
   fillChar(SetupParms,sizeof(SetupParms),0);
   with SetUpParms do begin
      ulBitsPerSample:=AudioBytes*8;
      ulFormatTag:=MCI_WAVE_FORMAT_PCM;
      ulSamplesPerSec:=AudioRate;
      ulChannels:=Channels;
      ulFormatMode:=MCI_RECORD;
      ulDeviceType:=MCI_DEVTYPE_WAVEFORM_AUDIO;
      pmixEvent:=Event;
   end;
   MMCheck(mciSendCommand(DartDevice,MCI_MIXSETUP,MCI_WAIT+MCI_MIXSETUP_INIT,SetUpParms,0));
   DartRead:=SetUpParms.pmixRead;
   MixHandle:=SetUpParms.ulMixHandle;
   fillChar(BufferParms,sizeof(BufferParms),0);
   fillChar(BufferList,sizeof(BufferList),0);
   AudioBufferSize:=(trunc(AudioRate/framerate)+1)*SampleSize;
   BufferFraction:=(AudioRate/Framerate)-int(AudioRate/FrameRate);
   with BufferParms do begin
      ulStructLength:=sizeof(BufferParms);
      ulNumBuffers:=MaxBuffers;
      ulBufferSize:=AudioBufferSize;
      pBufList:=@BufferList;
   end;
   MMCheck(mciSendCommand(DartDevice,MCI_BUFFER,MCI_WAIT+MCI_ALLOCATE_MEMORY,BufferParms,0));
   BufferCount:=BufferParms.ulNumBuffers;
   DosCreateEventSem(nil,sem,0,true);
   DosCreateMutexSem(nil,BuffReadyMutex,0,false);
end;

function StartRec(Parameter : Pointer) : LongInt;
var
   dop : DEVOPENSTRUC;
   dc : HDC;
   size : SIZEL;
   bmp : BITMAPINFOHEADER2;
   hbm : HBITMAP;
   ps : HPS;
   pbit,
   d,
   p : pbitmap;
   bmpInfo : BITMAPINFO2;
   aptl : array[0..2] of POINTL;
   hbmOld : HBITMAP;
   BTTVPS : HPS;
   s : string;
   f : hFILE;
   action : ULONG;
   bmprow,
   rawrow,
   bmpsize,
   rawsize : integer;
   loop : integer;
   baset,
   CurrNow,
   t : tDateTime;
   Multitrackwrite : MMMULTITRACKWRITE;
   MyTrackMap : array[0..1] of TrackMap;
   Aud,
   Vid : RECORDTABWRITE;
   filelen : integer;
   VideoCompressInfo : MMVIDEOCOMPRESS;
   CompressInfo : MMCOMPRESS;
   row : integer;
   FrameSWP,
   ClientSWP : SWP;
   CurrBuffersReady : integer;
   alData : array[0..1] of LONG;
   BuffBase : double;
begin
   framecount:=round(lengthsec*framerate);
{      WinQueryWindowPos(BTTVFrame,FrameSWP);
   WinQueryWindowPos(BTTVClient,ClientSWP);
   FrameSWP.cx:=width+FrameSWP.cx-ClientSWP.cx;
   FrameSWP.cy:=Height+FrameSWP.cy-ClientSWP.cy;
   FrameSWP.fl:=SWP_SIZE;
   WinSetWindowPos(BTTVFrame,HWND_TOP,0,0,width+FrameSWP.cx-ClientSWP.cx,Height+FrameSWP.cy-ClientSWP.cy,SWP_SIZE+SWP_ZORDER);
//      WinPostMsg(BTTVFrame,WM_ADJUSTWINDOWPOS,ULONG(@FrameSWP),0);}
   fillchar(Dop,sizeof(Dop),0);
   Dop.pszDriverName:='DISPLAY';
   dc:=DevOpenDC(ab,OD_MEMORY,'*',5,@dop,NULLHANDLE);
   size.cx:=width;
   size.cy:=Height;
   ps:=GpiCreatePS(ab,dc,size,PU_PELS+GPIA_ASSOC+GPIT_MICRO);
   gpiQueryDeviceBitmapFormats(ps,2,alData[0]);
   fillchar(bmp,sizeof(bmp),0);
   bmp.cbFix:=sizeof(bmp);
   bmp.cx:=width;
   bmp.cy:=Height;
   bmp.cPlanes:=1;
   BitCount:=alData[1];
   bmp.cBitCount:=BitCount;
   bmp.ulCompression:=BCA_UNCOMP;
   bmp.usUnits:=BRU_METRIC;
   bmp.usRecording:=BRA_BOTTOMUP;
   bmp.usRendering:=BRH_NOTHALFTONED;
   bmp.ulColorEncoding:=BCE_RGB;
   hbm:=GpiCreateBitmap(ps,bmp,0,bmpinfo,bmpinfo);
   hbmOld:=GpiSetBitmap(ps,hbm);
   GpiQueryBitmapDimension(hbm,size);
   fillchar(aptl,sizeof(aptl),0);
   aptl[1].x:=width;
   aptl[1].y:=Height;
   BTTVPS:=WinGetPS(BTTVClient);
   bmprow:=(((BitCount*width)+31) div 32)*4;
   rawrow:=bitcount*width div 8;
//   bmprow:=rawrow;
   rawsize:=rawrow*Height;
   bmpsize:=bmprow*Height;
   Getmem(pbit,bmpsize);
   Getmem(p,rawsize);
   Getmem(d,rawsize);
   fillchar(bmpinfo,sizeof(bmpinfo),0);
   bmpinfo.cbFix:=16;
   bmpinfo.cPlanes:=1;
   bmpinfo.cx:=width;
   bmpinfo.cy:=Height;
   bmpinfo.cBitCount:=BitCount;
   filenum:=1;
   SampleSize:=AudioBytes*Channels;
   OpenFile;
   OpenDart;
   with Multitrackwrite do begin
      ulFlags:=0;
      ulReserved:=0;
      if HasAudio then ulNumTracks:=2
      else ulNumTracks:=1;
      pTrackMapList:=@MyTrackMap;
   end;
   with MyTrackMap[0] do begin
      ulTrackId:=0;
      ulNumEntries:=1;
      pRecordTabList:=@Vid;
   end;
   fillchar(vid,sizeof(vid),0);
   with Vid do begin
      if Uncompress then
         pRecord:=p
      else
         pRecord:=d;
      ulLength:=rawsize;//CompressInfo.ulDstBufLen;
   end;
   if HasAudio then begin
      with MyTrackMap[1] do begin
         ulTrackId:=1;
         ulNumEntries:=1;
         pRecordTabList:=@Aud;
      end;
      fillchar(aud,sizeof(aud),0);
      Aud.ulLength:=AudioBufferSize;
   end;
   fillChar(VideoCompressInfo,sizeof(VideoCompressInfo),0);
   VideoCompressInfo.ulStructLen:=sizeof(VideoCompressInfo);
   with CompressInfo do begin
      ulStructLen:=sizeof(CompressInfo);
      pRunTimeInfo:=@VideoCompressInfo;
      ulSrcBufLen:=rawsize;
      pSrcBuf:=p;
      ulDstBufLen:=rawsize;
      pDstBuf:=d;
      ulFlags:=MMIO_IS_KEY_FRAME
   end;
   Vid.ulParm2:=0;
   filelen:=0;
   BuffersReady:=0;
   CurrBuffer:=0;
   DosSetPriority(PRTYS_THREAD,PRTYC_NOCHANGE,30,0);
   baset:=now;
   t:=baset;
   BuffBase:=0.0;
   for loop:=0 to BufferCount-1 do begin
      BuffBase:=BuffBase+BufferFraction;
      if BuffBase>=1.0 then begin
         BufferList[loop].ulBufferLength:=AudioBufferSize;
         BuffBase:=BuffBase-1.0;
      end else
         BufferList[loop].ulBufferLength:=AudioBufferSize-SampleSize;
   end;
   loop:=0;
   DartRead(MixHandle,BufferList[0],BufferCount);
   while (loop<=framecount) and not AbortRec do begin
      if BuffersReady=0 then begin
         DosResetEventSem(sem,action);
         blocking:=true;
         DosWaitEventSem(sem,-1);
      end;
      CurrBuffersReady:=BuffersReady;
      gpiBitBlt(ps,BTTVPS,3,aptl[0],ROP_SRCCOPY,BBO_IGNORE);
      if HasAudio then begin
         Aud.pRecord:=BufferList[CurrBuffer].pBuffer;
      end else begin
         for row:=1 to CurrBuffersReady do
            with BufferList[CurrBuffer] do begin
               mmioWrite(wavmmio,pBuffer,ulBufferLength);
               BuffBase:=BuffBase+BufferFraction;
               if BuffBase>=1.0 then begin
                  ulBufferLength:=AudioBufferSize;
                  BuffBase:=BuffBase-1.0;
               end else
                  ulBufferLength:=AudioBufferSize-SampleSize;
               DartRead(MixHandle,BufferList[CurrBuffer],1);
               CurrBuffer:=(CurrBuffer+1) mod BufferCount;
            end;
         DosRequestMutexSem(BuffReadyMutex,-1);
         dec(BuffersReady,CurrBuffersReady);
         DosReleaseMutexSem(BuffReadyMutex);
         Vid.ulParm2:=CurrBuffersReady-1;
         inc(loop,CurrBuffersReady);
      end;
      gpiQueryBitmapBits(ps,0,height,pbit^,bmpinfo);
      if NoFlip then
         for row:=0 to Height-1 do begin
            move(pbit^[row*bmprow],p^[row*rawrow],rawrow);
         end
      else
         for row:=0 to Height-1 do begin
            move(pbit^[row*bmprow],p^[(Height-row-1)*rawrow],rawrow);
         end;
      if not UnCompress then begin
         CompressInfo.ulDstBufLen:=rawsize;
         mmioCheck(mmioSendMessage(mmio,MMIOM_COMPRESS,long(@compressinfo),0),'MMIOM_COMPRESS');
         Vid.ulLength:=CompressInfo.ulDstBufLen;
      end;
      mmioCheck(mmioSendMessage(mmio,MMIOM_MULTITRACKWRITE,long(@Multitrackwrite),sizeof(Multitrackwrite)),'MMIOM_MULTITRACKWRITE');
{         CurrBuffer:=(CurrBuffer+1) mod BufferCount;
      dec(BuffersReady);
      DartRead(MixHandle,BufferList[CurrBuffer],1);
      Vid.ulParm2:=0;}
//         sleeplen:=round((t-now)*factor);
      //inc(loop);
      inc(filelen,Vid.ulLength);
      if filelen>2000000000 then begin
         mmioCHeck(mmioClose(mmio,0),'mmioClose');
         mmioClose(wavmmio,0);
         inc(filenum);
         OpenFile;
         filelen:=0;
      end;
      {t:=baset+loop*frameLen;
      CurrNow:=now;
      while t<Currnow do begin
         inc(loop);
         t:=baset+loop*frameLen;
//            sleeplen:=round((t-now)*factor);
         inc(vid.ulParm2);
      end;
//         while (t>now) do dossleep(0);
      if (t-CurrNow)>(framelen/4) then dossleep(round((t-currnow)*factor));}
   end;
   mmioCheck(mmioClose(mmio,0),'mmioClose');
   mmioClose(wavmmio,0);
{      s:='P6 '+IntToStr(Width)+' '+IntToStr(Height)+' 255 ';
   DosOpen('out.ppm',f,action,rawsize+Length(s),0,OPEN_ACTION_CREATE_IF_NEW+OPEN_ACTION_REPLACE_IF_EXISTS,OPEN_SHARE_DENYWRITE+OPEN_FLAGS_SEQUENTIAL+OPEN_ACCESS_WRITEONLY,nil);
   DosWrite(f,s[1],length(s),action);
   DosWrite(f,p^,rawsize,action);
   DosClose(f);}
   freemem(p,rawsize);
   freemem(pbit,bmpsize);
   freemem(d,rawsize);
   GpiSetBitmap(ps,hbmOld);
   gpideletebitmap(hbm);
   WinReleasePS(BTTVPS);
   GPIDestroyPS(BTTVPS);
   DevCloseDC(dc);
   DosCloseMutexSem(BuffReadyMutex);
   DosCloseEventSem(sem);
   WinPostMsg(Client,WM_CLOSE,0,0);
end;

function CaptRecWndProc(Window : HWnd; Msg : ULong; MP1, MP2 : MParam): MResult; CDECL;
var
   ps : hps;
   PaintArea : RECTL;
   enum : HENUM;
   title : string;
   found : boolean;
   WinSize : swp;
begin
   result:=0;
   case Msg of
   WM_CREATE: begin
      Client:=Window;
      AbortRec:=false;
      enum:=WinBeginEnumWindows(HWND_DESKTOP);
      found:=false;
      BTTVFrame:=WinGetNextWindow(enum);
      while (BTTVFrame<>NULLHANDLE) and (not found) do begin
         SetLength(title,256);
         SetLength(title,WinQueryWindowText(BTTVFrame,256,pchar(title)));
         found:=title=WindowTitle;
         if not found then BTTVFrame:=WinGetNextWindow(enum);
      end;
      WinEndEnumWindows(enum);
      if FrameWindow then BTTVClient:=BTTVFrame
      else begin
         enum:=WinBeginEnumWindows(BTTVFrame);
         found:=false;
         BTTVClient:=WinGetNextWindow(enum);
         while (BTTVClient<>NULLHANDLE) and (not found) do begin
            found:=WinQueryWindowUShort(BTTVClient,QWS_ID)=FID_CLIENT;
            if not found then BTTVClient:=WinGetNextWindow(enum);
         end;
         WinEndEnumWindows(enum);
      end;
      WinQueryWindowPos(bttvClient,WinSize);
      width:=WinSize.cx;
      Height:=WinSize.cy;
      ThreadId:=VPBeginThread(StartRec,64*1024,nil);
      result:=WinDefWindowProc(Window,Msg,MP1,MP2);
   end;
   WM_PAINT : begin
      ps:=WinBeginPaint(Client,NullHandle,@PaintArea);
      WinFillRect(ps,PaintArea,SYSCLR_DIALOGBACKGROUND);
      WinEndPaint(Client);
   end;
   WM_CLOSE : begin
      AbortRec:=true;
      DosWaitThread(ThreadId,DCWW_WAIT);
      result:=WinDefWindowProc(Window,Msg,MP1,MP2);
   end;
   else result:=WinDefWindowProc(Window,Msg,MP1,MP2);
   end
end;

var
   Msg : QMsg;
   CreateFlags : ULONG;
   loop : integer;
   CurrParam : string;

begin
   {assign(log,'captrec.log');
   rewrite(log);}
   loop:=1;
   lengthsec:=0;
   BaseName:='out';
   Quality:=-1;
   framerate:=30;
   AudioBytes:=2;
   AudioRate:=22050;
   Channels:=2;
   HasAudio:=false;
   WindowTitle:='BTTV';
   CODECDLL:='ULCORT';
   CODECEntry:='CodecEntry';
   NoFlip:=false;
   FrameWindow:=false;
   for loop:=1 to paramcount do begin
      CurrParam:=Paramstr(loop);
      if CurrParam[1]='/' then begin
         //if not (CurrParam[2] in ('O','8','A')) and CurrParam[3]<>':' then raise Exception.Create('Value required for parameter /'+CurrParam[2]);
         case UpCase(CurrParam[2]) of
         'S' : lengthsec:=lengthsec+StrToInt(copy(CurrParam,4,255));
         'M' : lengthsec:=lengthsec+60*StrToInt(copy(CurrParam,4,255));
         'D' : CODECDLL:=copy(CurrParam,4,255);
         'E' : CODECEntry:=copy(CurrParam,4,255);
{         'H' : Height:=StrToInt(copy(CurrParam,4,255));
         'W' : begin
            width:=StrToInt(copy(CurrParam,4,255));
            Height:=width*3 div 4;
         end;}
         'Q' : Quality:=StrToInt(copy(CurrParam,4,255));
         'F' : FrameRate:=StrToFloat(copy(CurrParam,4,255));
         'R' : AudioRate:=StrToInt(copy(CurrParam,4,255));
         'O' : Channels:=1;
         '8' : AudioBytes:=1;
         //'A' : HasAudio:=true;
         'T' : WindowTitle:=copy(CurrParam,4,255);
         'L' : NoFlip:=true;
         'W' : FrameWindow:=true;
         else raise Exception.Create('Unknown parameter /'+CurrParam[2]);
         end
      end else basename:=CurrParam;
   end;
   framelen:= (1/framerate)/(60*60*24);
   UnCompress:=Uppercase(CODECDLL)='MONDO';
   if UpperCase(copy(BaseName,length(BaseName)-3,255))='.AVI' then setlength(basename,length(basename)-4);
   ab:=WinInitialize(0);
   mq:=WinCreateMsgQueue(ab,0);
   WinRegisterClass(ab,'captrec',CaptRecWndProc,CS_SIZEREDRAW,0);
   Createflags:=FCF_TITLEBAR+FCF_SYSMENU+FCF_SIZEBORDER+FCF_MINBUTTON+FCF_SHELLPOSITION+FCF_TASKLIST;
   Frame:=WinCreateStdWindow(HWND_DESKTOP,WS_VISIBLE,CreateFlags,'captrec','Capture Record',WS_CLIPCHILDREN,0,100,@Client);
   WinSetWindowPos(Frame,0,0,0,100,100,SWP_SIZE);
   while WinGetMsg(ab,Msg,0,0,0) do WinDispatchMsg(ab,Msg);
   WinDestroyWindow(Frame);
   WinDestroyMsgQueue(mq);
   WinTerminate(ab);
   {close(log);}
end.
